/*
	vmsvga.c	screen driver
        video board / chip dependent processing: VMware SVGA II

	Copyright 2015 by SASANO Takayoshi
	This software is distributed under the T-License 2.0.
*/
#include "screen.h"
#include <device/videomode.h>
#include <tk/util.h>
#include "svga.h"

#define	VIDEOMODE	DM2048x32
#define	SUPPORT_MODEMAP	(1 << VIDEOMODE)

LOCAL	FastLock	SVGALock;

/* update region */
LOCAL	void	VMSVGAupdate(W x, W y, W dx, W dy)
{
	Lock(&SVGALock);
	SVGA_Update(x, y, dx, dy);
	Unlock(&SVGALock);

	return;
}

/* set color map (not supported) */
LOCAL	void	VMSVGAsetcmap(COLOR *cmap, W index, W entries)
{
	/* no support, do nothing */
	return;
}

/* set display mode */
LOCAL	void	VMSVGAsetmode(W flg)
{
	/* exit VMware SVGA II mode, required for warm reboot */
	// XXX the last contents of VGA mode is redisplayed when exiting.
	if (flg < 0) {
		Lock(&SVGALock);
		SVGA_Disable();
		/* exit with lock */
		return;
	}

	/* enter VMware SVGA II mode */
	SVGA_SetMode(Vinf.act_width, Vinf.act_height, Vinf.pixbits >> 8,
		     (Vinf.attr & USE_VVRAM) ? TRUE : FALSE);

	/* fix display mode information */
	Vinf.width = Vinf.fb_width = Vinf.act_width;
	Vinf.height = Vinf.fb_height = Vinf.act_height;
	Vinf.rowbytes = Vinf.framebuf_rowb = gSVGA->pitch;
	Vinf.vramsz = Vinf.framebuf_rowb * Vinf.fb_height;

	return;
}

/* suspend / resume processing */
LOCAL	void	VMSVGAsuspend(BOOL suspend)
{
        /* in the case of suspend, clear Video-RAM content */
	if (suspend) {
		Lock(&SVGALock);
		memset(Vinf.f_addr, 0, Vinf.framebuf_total);
		if (Vinf.attr & USE_VVRAM) {
			SVGA_Update(0, 0, Vinf.width, Vinf.height);
		}
	} else {
		Unlock(&SVGALock);
	}

	return;
}

/* initialization processing */
EXPORT	W	VMSVGAInit(void)
{
	ER	er;
	T_CFLG	cflg = {
		.flgatr = TA_TFIFO,
		.iflgptn = 0,
	};

	/* allocate resources for VMware-SVGA II driver */
	gSVGA = Kcalloc(1, sizeof(SVGADevice));
	if (gSVGA == NULL) {
		er = E_NOMEM;
		goto fin0;
	}

	er = tk_cre_flg(&cflg);
	if (er < E_OK) goto fin1;
	gSVGA->FlagId = er;

	er = CreateLock(&SVGALock, "vmsc");
	if (er < E_OK) goto fin2;

	/* check device */
	if (SVGA_Init()) return E_NOMDA;

	/* set Vinf */
	Vinf.reqmode = Vinf.curmode = VIDEOMODE;
	Vinf.framebuf_addr = (void *)gSVGA->fbBase;
	Vinf.f_addr = gSVGA->fbMem;
	strncpy(Vinf.chipinf, "VMware SVGA II", L_CHIPINF);
	Vinf.framebuf_total = gSVGA->fbSize;
	Vinf.attr |= (LINEAR_FRAMEBUF | NEED_FINPROC | NEED_SUSRESPROC);
	Vinf.attr &= ~BPP_24;
	Vinf.fn_setcmap = VMSVGAsetcmap;
	Vinf.fn_setmode = VMSVGAsetmode;
	Vinf.fn_susres = VMSVGAsuspend;
	Vinf.modemap = SUPPORT_MODEMAP;

	if (Vinf.attr & USE_VVRAM) {
		Vinf.fn_updscr = VMSVGAupdate;
		Vinf.v_addr = gSVGA->fbMem;
	}

	/* these values are temporally, fix them at VMSVGAsetmode() */
	Vinf.fb_width = Vinf.framebuf_rowb = VideoHsize(Vinf.reqmode);
	Vinf.fb_height = VideoVsize(Vinf.reqmode);

	er = 1;
	goto fin0;

/*fin3:*/
	DeleteLock(&SVGALock);
fin2:
	tk_del_flg(gSVGA->FlagId);
fin1:
	Kfree(gSVGA);
fin0:
	return er;
}
